@@ -15,27 +15,21 @@
 import __main__
 import threading, os, sys, re
 import paramiko
-from paramiko.common import (
-    DEBUG,
-    ERROR,
-)
+from logging import DEBUG
 import time
 import subprocess
 from select import select
 import errno
 import asyncio
 try:
-    from sshscripterror import SSHScriptError
+    from .sshscripterror import SSHScriptError, getLogger
 except ImportError:
-    from .sshscripterror import SSHScriptError
-
-#global logger
-#logger = SSHScriptDummyLogger
+    from sshscripterror import SSHScriptError, getLogger
 
 class WithChannelWrapper(object):
     def __init__(self,channel):
         self.c = channel
-    def sendline(self,s='\n',secondsToWaitResponse=1):
+    def sendline(self,s='\n',secondsToWaitResponse=None):
         return self.c.sendline(s,secondsToWaitResponse)
     def expect(self,rawpat,timeout=60,stdout=True,stderr=True):
         return self.c.expect(rawpat,timeout,stdout,stderr)
@@ -47,19 +41,21 @@ class WithChannelWrapper(object):
         return self.c.wait(seconds)
 
     @property
+    def channel(self):
+        return self.c.channel
+    @property
     def stdout(self):
         return self.c.stdout
     @property
     def stderr(self):
         return self.c.stderr
+    
 
 
 class GenericChannel(object):
+    
     def __init__(self):
-        #global logger
-        #if logger is SSHScriptDummyLogger:
-        #    logger = __main__.SSHScript.logger
-        self.logger = paramiko.util.get_logger('sshscript')
+        self.logger = getLogger()
         self.allStdoutBuf = []
         self.allStderrBuf = [] 
         self._stdout = ''
@@ -67,21 +63,13 @@ class GenericChannel(object):
         self.stdoutBuf = []
         self.stderrBuf = []
         self.lock = threading.Lock()
-        # 不要0，這樣可以讓一開始就先等一等
-        self._lastIOTime = time.time()
 
         # verbose-related
         self.stdoutDumpBuf = []
         self.stderrDumpBuf = []
-        try:
-            self.loop = asyncio.get_event_loop()
-        except RuntimeError:
-            self.loop = asyncio.new_event_loop()
-            asyncio.set_event_loop(self.loop)
         
         if os.environ.get('VERBOSE'):
             self.dump2sys = True
-            #🔵🔴🟠🟡🟢🟣🟤⭕⬜⬛🔲🟦🟥🟧🟨🟩🟪🟫🛑🔶🔷🔸🔹🔺🔻 
             self.stdoutPrefix = os.environ.get('VERBOSE_STDOUT_PREFIX','🟩').encode('utf8')
             self.stderrPrefix = os.environ.get('VERBOSE_STDERR_PREFIX','🟨').encode('utf8')
         else:
@@ -123,13 +111,25 @@ class GenericChannel(object):
             self.close()
             raise SSHScriptError(self.stderr)
  
-    async def waitio(self,waitingInterval=None):
+    async def waitio_old(self,waitingInterval=None):
         if waitingInterval is None:
             waitingInterval = float(os.environ.get('CMD_INTERVAL',0.5))
+        '''
         remain = waitingInterval - (time.time() - self._lastIOTime)
         while remain > 0:
             await asyncio.sleep(remain)    
             remain = waitingInterval - (time.time() - self._lastIOTime)
+        '''
+        remain = waitingInterval - (time.time() - self.LastIOTime)
+        while remain > 0:
+            await asyncio.sleep(remain)    
+            remain = waitingInterval - (time.time() - self.LastIOTime)
+        #如果有螢幕輸出的話，幫助它順序維持正確的順序
+        sys.stdout.flush()
+        sys.stderr.flush()
+
+    async def waitio(self,waitingInterval):
+        await self.owner.sshscript.waitio(waitingInterval)
         #如果有螢幕輸出的話，幫助它順序維持正確的順序
         sys.stdout.flush()
         sys.stderr.flush()
@@ -171,19 +171,15 @@ class GenericChannel(object):
                     raise TimeoutError(f'Not found: {pat} ' + '\n')
                 for dataSource in targets:
                     di = dataSource()
-                    for pat in pats:
+                    for idx,pat in enumerate(pats):
                         if pat.search(di):
                             await self.waitio(1)
-                            return
+                            return rawpat[idx]
                 await asyncio.sleep(0.1)
         return self.loop.run_until_complete(_wait())    
         
-    def wait(self,seconds=None):
-        async def _wait(seconds):
-            await self.waitio(seconds)
-        return self.loop.run_until_complete(_wait(seconds))    
-        
-    def wait(self,seconds=None):
+       
+    def wait(self,seconds):
         async def _wait(seconds):
             await self.waitio(seconds)
         return self.loop.run_until_complete(_wait(seconds))                
@@ -195,21 +191,25 @@ class GenericChannel(object):
         self.wait(1)
         self.close()
     
-    def sendline(self,s='\n',secondsToWaitResponse=1):
-        #global logger
-        self._log(DEBUG,f'sendline: {s}')
-        if not s[-1] == '\n': s += '\n'
-        # 確保跟前面的一個指令有點「距離」，不要在還在接收資料時送出下一個指令
-        self.wait()
-        n = self.send(s)
-        if secondsToWaitResponse > 0:
-            self.recv(secondsToWaitResponse)
-        return n
+    def sendline(self,line='',waitingInterval=None):
+        # accept multiple lines from v1.1.13
+        if waitingInterval is None:
+            waitingInterval = self.owner.waitingInterval
+        lines = [x.lstrip() for x in line.splitlines()]
+        for idx,line in enumerate(lines):
+            if not line: line = '\n'
+            elif not line[-1] == '\n': line += '\n'
+            # 確保跟前面的一個指令有點「距離」，不要在還在接收資料時送出下一個指令
+            self.send(line)
+            self._log(DEBUG,f'[sshscriptchannel]sendline({waitingInterval}):{line[:-1]}')
+            # 在sshscriptdollar中呼叫前已經先wait過，因此很多行時，先送再wait
+            self.wait(waitingInterval)
+        self.recv(waitingInterval)
 
     def addStdoutData(self,x):  
         # 不要讓self.lock.acquire()影響self._lastIOTime
-        self._lastIOTime = time.time()
         if not x:return
+        self.owner.sshscript.touchIO('addStdoutData')
         self.lock.acquire()
         self.stdoutBuf.append(x)
         self.allStdoutBuf.append(x)
@@ -231,8 +231,8 @@ class GenericChannel(object):
     
     def addStderrData(self,x):
         # 不要讓self.lock.acquire()影響self._lastIOTime
-        self._lastIOTime = time.time()
         if not x: return
+        self.owner.sshscript.touchIO('addStderrData')        
         self.lock.acquire()
         self.stderrBuf.append(x)
         self.allStderrBuf.append(x)
@@ -251,91 +251,18 @@ class GenericChannel(object):
             for line in lines:
                 sys.stderr.buffer.write(self.stderrPrefix + line + b'\n')
             sys.stderr.buffer.flush()
-'''
-class POpenPipeChannel(GenericChannel):
-    def __init__(self,owner,cp,timeout):
-        super().__init__()
-        self.owner = owner
-        self.cp = cp
-        if cp is None:
-            # dummy instance for "with $<command> as ..."
-            self._lastIOTime = 0
-        else:
-            self.masterFd = [cp.stdout,cp.stderr]
-            self.slaveFd = [cp.stdin]
-            self.timeout = timeout
-            threading.Thread(target=self._reading).start()
 
-    def send(self,s):
-        self._lastIOTime = time.time()
-        b = s.encode('utf-8')
-        # write to popen's stdin
-        self.slaveFd[0].write(b)
-        self.slaveFd[0].flush()
-        self._lastIOTime = time.time()
-    
-    def _reading(self):
-
-        buffer = {
-            self.masterFd[0]: self.addStdoutData, #stdout
-            self.masterFd[1]: self.addStderrData  #stderr
-        }
-        while buffer :           
-            try:
-                for fd in select(buffer, [], [],1)[0]:
-                    try:
-                        # diff of read and read1:
-                        # See: https://stackoverflow.com/questions/57726771/what-the-difference-between-read-and-read1-in-python
-                        data = fd.read1(1024)
-                    except OSError as e:
-                        if e.errno == errno.EIO:
-                            pass
-                        elif e.errno == errno.EBADF:
-                            pass
-                        else:
-                            raise #XXX cleanup
-                        del buffer[fd] # EIO means EOF on some systems
-                    else:
-                        if not data: # EOF
-                            del buffer[fd]
-                        else:
-                            self._lastIOTime = time.time()
-                            buffer[fd](data)
-                            
-            except OSError as e:
-                if e.errno == errno.EBADF:
-                    # being closed
-                    break
-                else:
-                    raise
-
-    def close(self):
-        self.wait(1) # at least 1 seconds has no io
-        if self.cp: # not dummy instance
-            self.masterFd[0].close()
-            self.masterFd[1].close()
-            self.slaveFd[0].close()
-            self.cp.stdin.close()
-
-            while self.cp.poll() is None:
-                try:
-                    self.cp.wait(self.timeout)
-                except subprocess.TimeoutExpired:
-                    self.cp.kill()
-                    break
-    
-        self.owner.stderr = (b''.join(self.allStderrBuf)).decode('utf8')
-        self.owner.stdout = (b''.join(self.allStdoutBuf)).decode('utf8')    
-'''
 class POpenChannel(GenericChannel):
     def __init__(self,owner,cp,timeout,masterFd=None,slaveFd=None):
         super().__init__()
         self.owner = owner
+        self.loop = self.owner.sshscript.loop
         self.closed = False
         self.cp = cp
         if cp is None:
             # dummy instance for "with $<command> as ..."
-            self._lastIOTime = 0
+            #self._lastIOTime = 0
+            pass
         else:
             self.masterFd = masterFd
             self.slaveFd = slaveFd
@@ -343,15 +270,12 @@ class POpenChannel(GenericChannel):
             threading.Thread(target=self._reading).start()
 
     def send(self,s):
-        #self._lastIOTime = time.time()
         b = s.encode('utf-8')
-        # write to popen's stdin
-        #os.write(self.masterFd[0],b)
-        #os.fsync(self.masterFd[0])
+        self.lock.acquire()
         self.cp.stdin.write(b)
         self.cp.stdin.flush()
-        self._lastIOTime = time.time()
-    
+        self.lock.release()
+        self.owner.sshscript.touchIO('subprocess send '+s[:-1])    
     def _reading(self):
 
         buffer = {
@@ -377,7 +301,6 @@ class POpenChannel(GenericChannel):
                         if not data: # EOF
                             del buffer[fd]
                         else:
-                            self._lastIOTime = time.time()
                             buffer[fd](data)
                             
             except OSError as e:
@@ -390,29 +313,41 @@ class POpenChannel(GenericChannel):
     def close(self):
         self.closed = True
         self.wait(1) # at least 1 seconds has no io
+        error = None
         if self.cp: # not dummy instance
             os.close(self.masterFd[0])
             os.close(self.masterFd[1])
             os.close(self.slaveFd[0])
             os.close(self.slaveFd[1])
             self.cp.stdin.close()
+            '''
             while self.cp.poll() is None:
                 try:
                     self.cp.wait(self.timeout)
                 except subprocess.TimeoutExpired:
                     self.cp.kill()
                     break
-    
+            self.owner.exitcode = self.cp.returncode
+            '''
+            try:
+                self.owner.exitcode = self.cp.wait(self.timeout)
+            except subprocess.TimeoutExpired as e:
+                error = e
+                self.cp.kill()
+            self._log(DEBUG,f'[subprocess] exitcode={self.owner.exitcode}')
+                
+
         self.owner.stderr = (b''.join(self.allStderrBuf)).decode('utf8',errors='ignore')
         self.owner.stdout = (b''.join(self.allStdoutBuf)).decode('utf8',errors='ignore')    
-
+        if error: raise error
 
 class ParamikoChannel(GenericChannel):
     def __init__(self,owner,client,timeout):
         super().__init__()        
         self.owner = owner # an instance of SSHScriptDollar
-        self.client = client
-        if self.client:
+        self.loop = self.owner.sshscript.loop
+        if isinstance(client,paramiko.client.SSHClient):
+            self.client = client
             self.timeout = timeout
             self.channel = client.invoke_shell(term='vt100') # a paramiko.Channel
             self.stdin = self.channel.makefile_stdin('wb',-1)
@@ -420,30 +355,56 @@ class ParamikoChannel(GenericChannel):
             self.stderrStream = self.channel.makefile_stderr('r', -1) 
             self.channel.settimeout(timeout)
             threading.Thread(target=self._reading).start()
+        elif isinstance(client,paramiko.channel.Channel):
+            self.client = None
+            self.channel = client
+            self.timeout = timeout
+            self.stdin = self.channel.makefile_stdin('wb',-1)
+            self.stdoutStream = self.channel.makefile('rb', -1)
+            self.stderrStream = self.channel.makefile_stderr('r', -1) 
+            self.channel.settimeout(timeout)
+            threading.Thread(target=self._reading).start()
+        else:
+            # dummy
+            self.client = None
+            self.channel = None
 
     def send(self,s):
         while not self.channel.send_ready():
-            time.sleep(0.1)
-        self._lastIOTime = time.time()
+            time.sleep(0.05)
+        '''
+        既然要raise,就讓paramiko raise就好
         n = self.channel.sendall(s)
         if n == 0:
             raise IOError('failed to write to channel')
-        self._lastIOTime = time.time()
+        '''
+        
+        self.lock.acquire()
+        n = self.channel.sendall(s)
+        self.lock.release()
+        self.owner.sshscript.touchIO('ssh send')
         return n
             
     def _reading(self):
-    
-        while not self.channel.closed:
-            if self.channel.recv_ready():
-                data = self.channel.recv(1024)
-                self.addStdoutData(data)
-            if self.channel.recv_stderr_ready():
+        # this is run in thread
+        while not (self.channel.closed or self.channel.exit_status_ready()):
+            while self.channel.recv_stderr_ready():
                 data = self.channel.recv_stderr(1024)
                 self.addStderrData(data)
+            while self.channel.recv_ready():
+                data = self.channel.recv(1024)
+                self.addStdoutData(data)
     
+        while not (self.channel.exit_status_ready()):
+            time.sleep(0.1)
+        self.owner.exitcode = self.channel.recv_exit_status()
+        self._log(DEBUG,f'[ssh] exit code= {self.owner.exitcode}')
+
     def close(self):
-        if self.client: # not dummy instance
+        if self.channel: # not dummy instance
+            self.channel.shutdown_write()
             self.channel.close()
+
         # 此時呼叫self.recv()沒有意義，只要呼叫wait()就好
         self.wait(1)
         self.owner.stderr = (b''.join(self.allStderrBuf)).decode('utf8',errors='ignore')
